/*
 *  Copyright 1996, 1997, 1998 Hans Reiser, see reiserfs/README for licensing and copyright details
 */


/*
 * Contains code from
 *
 *  linux/include/linux/lock.h and linux/fs/buffer.c /linux/fs/minix/fsync.c
 *
 *  Copyright (C) 1991, 1992  Linus Torvalds
 */
#ifdef __KERNEL__

#include <linux/sched.h>
#include <linux/locks.h>
#include <linux/reiserfs_fs.h>
#include <linux/smp_lock.h>



/*
 *  wait_buffer_until_released
 *  reiserfs_bread
 *  reiserfs_getblk
 *  reiserfs_sync_block
 *  sync_unf_nodes
 *  sync_file_item
 *  sync_file_items
 *  reiserfs_sync_file
 */


/* when we allocate a new block (get_new_buffer, get_empty_nodes,
   get_nodes_for_preserving) and get buffer for it, it is possible
   that it is held by someone else or even by this process. In this
   function we wait until all other holders release buffer. To make
   sure, that current process does not hold we did free all buffers in
   tree balance structure (get_empty_nodes and
   get_nodes_for_preserving) or in path structure only
   (get_new_buffer) just before calling this */
void wait_buffer_until_released (struct buffer_head * bh)
{
  int repeat_counter = 0;

  while (atomic_read (&(bh->b_count)) > 1) {
    if ( !(++repeat_counter % 2000000) ) {
      reiserfs_warning ("vs-3050: wait_buffer_until_released: nobody releases buffer (%b). Still waiting (%d) %cJDIRTY %cJWAIT\n",
			bh, repeat_counter, buffer_journaled(bh) ? ' ' : '!',
			buffer_journal_dirty(bh) ? ' ' : '!');
    }
    run_task_queue(&tq_disk);
    current->policy |= SCHED_YIELD;
    /*current->counter = 0;*/
    schedule();
  }
}


/* no longer need, should just make journal.c use the default handler */
void reiserfs_journal_end_io (struct buffer_head *bh, int uptodate)
{
  mark_buffer_uptodate(bh, uptodate);
  unlock_buffer(bh);
  return ;
}


/* struct used to service end_io events.  kmalloc'd in 
** reiserfs_end_buffer_io_sync 
*/
struct reiserfs_end_io {
  struct buffer_head *bh ; /* buffer head to check */
  struct tq_struct task ;  /* task struct to use */
  struct reiserfs_end_io *self ; /* pointer to this struct for kfree to use */
} ;

/*
** does the hash list updating required to release a buffer head.
** must not be called at interrupt time (so I can use the non irq masking 
** spinlocks).  Right now, put onto the schedule task queue, one for
** each block that gets written
*/
static void reiserfs_end_io_task(struct reiserfs_end_io *io) {
  struct buffer_head *bh = io->bh ;
  int windex = push_journal_writer("end_io_task") ;

  lock_kernel() ;
  if (buffer_journal_dirty(bh)) {
    struct reiserfs_journal_cnode *cur ;
    struct super_block * s = get_super (bh->b_dev);

    if (!s) 
      goto done ;

    if (!buffer_journal_dirty(bh)) { 
      goto done ;
    }
    mark_buffer_notjournal_dirty(bh) ;
    cur = (journal_hash(SB_JOURNAL(s)->j_list_hash_table, bh->b_dev, bh->b_blocknr)) ;
    while(cur) {
      if (cur->bh && cur->blocknr == bh->b_blocknr && cur->dev == bh->b_dev) {
	if (cur->jlist) { /* since we are clearing the bh, we must decrement nonzerolen */
	  atomic_dec(&(cur->jlist->j_nonzerolen)) ;
	}
	cur->bh = NULL ;
      }
      cur = cur->hnext ;
    }
    atomic_dec(&(bh->b_count)) ;
  }
done:
  kfree(io->self) ;
  pop_journal_writer(windex) ;
  unlock_kernel() ;
  return ;
}

/*
** general end_io routine for all reiserfs blocks.
** logged blocks will come in here marked buffer_journal_dirty()
** a reiserfs_end_io struct is kmalloc'd for them, and a task is put 
** on the scheduler queue.  It then does all the required hash table
** operations to reflect the buffer as writen
*/
void reiserfs_end_buffer_io_sync (struct buffer_head *bh, int uptodate)
{

  mark_buffer_notjournal_new(bh) ;
  if (buffer_journal_dirty(bh)) {
    struct reiserfs_end_io *io = kmalloc(sizeof(struct reiserfs_end_io), 
                                         GFP_ATOMIC) ;
    /* note, if kmalloc fails, this buffer will be taken care of
    ** by a check at the end of do_journal_end() in journal.c
    */
    if (io) {
      io->task.next = NULL ;
      io->task.sync = 0 ;
      io->task.routine = (void *)(void *)reiserfs_end_io_task ;
      io->task.data = io ;
      io->self = io ;
      io->bh = bh ;
      queue_task(&(io->task), &tq_scheduler) ;
    } else {
      printk("reiserfs/buffer.c-184: kmalloc returned NULL block %lu\n", 
              bh->b_blocknr) ;
    }
  }
  mark_buffer_uptodate(bh, uptodate);
  unlock_buffer(bh);
}


/*
 * reiserfs_bread() reads a specified block and returns the buffer that contains
 * it. It returns NULL if the block was unreadable.
 */
/* It first tries to find the block in cache, and if it cannot do so
   then it creates a new buffer and schedules I/O to read the
   block. */
/* The function is NOT SCHEDULE-SAFE! */

struct buffer_head  * reiserfs_bread (kdev_t n_dev, int n_block, int n_size) 
{
    struct buffer_head * bh;

    bh = bread (n_dev, n_block, n_size);
    if (bh) {
        bh->b_end_io = reiserfs_end_buffer_io_sync;
    }
    return bh;
}


/* This function looks for a buffer which contains a given block.  If
   the block is in cache it returns it, otherwise it returns a new
   buffer which is not uptodate.  This is called by reiserfs_bread and
   other functions. Note that get_new_buffer ought to be called this
   and this ought to be called get_new_buffer, since this doesn't
   actually get the block off of the disk. */
/* The function is NOT SCHEDULE-SAFE! */

struct buffer_head  * reiserfs_getblk (kdev_t n_dev, int n_block, int n_size)
{
    struct buffer_head * bh;

    bh = getblk (n_dev, n_block, n_size);
    if (bh) {
      bh->b_end_io = reiserfs_end_buffer_io_sync ;
    }
    return bh;
}

/* most of the code for syncing files is not needed in 2.3.  This #if 0
** ends right before the last function in this file
*/
#if 0

/* Synchronize a block of reiserfs. */
static int  reiserfs_sync_block(
	      struct reiserfs_transaction_handle *th, /* pointer to the transaction handle */
              struct buffer_head  * p_s_bh,   /* Pointer to the buffer header to sync.  */
              int                   n_wait,   /* Wait parameter.                        */
              struct path         * p_s_path  /* Pointer to the path contains buffer.
                                                  NULL in case of an unformatted node.  */
            ) {
    struct super_block *s ;
    return 0;

  if ( n_wait && buffer_req(p_s_bh) && ! buffer_uptodate(p_s_bh) )  {
    /* Release buffer if it is unformatted node. If not it will be released by pathrelse(). */
    if ( ! p_s_path )
      brelse(p_s_bh);
    return -1;
  }
  if ( n_wait || ! buffer_uptodate(p_s_bh) || ! buffer_dirty(p_s_bh) )  {
    if ( ! p_s_path )
      brelse(p_s_bh);
    return 0;
  }

  if (p_s_path) {
    s = th->t_super ;
    journal_mark_dirty(th, s, p_s_bh) ;
  } else { /* unformatted nodes can go right to the disk, BUG, except when they are part of a direct->indirect conversion */
    ll_rw_block(WRITE, 1, &p_s_bh); 
  }
  atomic_dec (&(p_s_bh->b_count));

  /* Decrement path length if it is buffer at the path. */
  if ( p_s_path ) {

#ifdef REISERFS_CHECK
    if ( p_s_path->path_length < FIRST_PATH_ELEMENT_OFFSET )
      reiserfs_panic(0, "PAP-16015: reiserfs_sync_block: path length is too small");
#endif

    p_s_path->path_length--;
  }

  return 0;
}


/* Sync unformatted nodes of the item *p_s_ih. */
static int  sync_unf_nodes(
	      struct reiserfs_transaction_handle *th,
              struct inode      * p_s_inode,        /* Pointer to the file inode.         */
              struct item_head  * p_s_ih,           /* Pointer to the found item header.  */
              int               * p_n_pos_in_item,  /* Position in the found item.        */
              struct path       * p_s_path_to_item, /* Path to the found item.            */
              int                 n_wait,           /* Wait parameter.                    */
              unsigned long     * p_n_synced  /* Returned value. UNFM number were
                                                        synced in this call.        */
            ) {

  return 0;

#if 0
  struct cpu_key            s_item_key;
  struct buffer_head  * p_s_unfm_bh;
  unsigned long         n_unfm_pointer;
  int                   n_counter,
                        n_unfm_number_to_sync = I_UNFM_NUM(p_s_ih) - *p_n_pos_in_item;

  /* Form key to search for the first unformatted node to be synced. */
  copy_key(&s_item_key, &(p_s_ih->ih_key));
  set_cpu_key_k_offset (&s_item_key, cpu_key_k_offset (&s_item_key) + *p_n_pos_in_item * p_s_inode->i_sb->s_blocksize);
  /*k_offset (&s_item_key) += *p_n_pos_in_item * p_s_inode->i_sb->s_blocksize;*/
  if ( search_for_position_by_key(p_s_inode->i_sb, &s_item_key, p_s_path_to_item) == POSITION_NOT_FOUND )
    return 1; /* Item to sync UNFM was not found. */
  /* Remember found item header.  */
  copy_item_head(p_s_ih, PATH_PITEM_HEAD(p_s_path_to_item));

  /* Calculate number of the UNFM to be synced. */
  if ( n_unfm_number_to_sync > I_UNFM_NUM(p_s_ih) - *p_n_pos_in_item/*vs*/)
    /* do not sync more unformatted nodes, than number of synced unformatted node pointers. If there
       is no so many unformatted nodes in found indirect item as we need, sync what we have. */
    n_unfm_number_to_sync = I_UNFM_NUM(p_s_ih) - *p_n_pos_in_item/*vs*/;

  for ( n_counter = 0; n_counter < n_unfm_number_to_sync; n_counter++ ) {
    /* Found item is not at the path. */
    if ( comp_items(p_s_ih, p_s_path_to_item) )
      return 0;
    /* Calculate block number of the UNFM. */
    n_unfm_pointer = B_I_POS_UNFM_POINTER(PATH_PLAST_BUFFER(p_s_path_to_item),
					  p_s_ih, *p_n_pos_in_item + n_counter);
    /* It is hole. Nothing to sync. */
    if ( ! n_unfm_pointer ) {
      *p_n_synced += 1;
      continue;
    }
    /* Get buffer contains UNFM. */
    p_s_unfm_bh = get_hash_table(p_s_inode->i_dev, n_unfm_pointer, p_s_inode->i_sb->s_blocksize);
    /* There is not needed buffer. */
    if ( ! p_s_unfm_bh )  {
      *p_n_synced += 1;
      continue;
    }

    if ( buffer_locked (p_s_unfm_bh) ) {
      __wait_on_buffer(p_s_unfm_bh);

      /* Check whether the found item at the path. */
      if ( comp_items(p_s_ih, p_s_path_to_item) ) {
	brelse(p_s_unfm_bh);
	return 0;
      }
      /* Number of the UNFM is changed. */
      if ( n_unfm_pointer != B_I_POS_UNFM_POINTER(PATH_PLAST_BUFFER(p_s_path_to_item), p_s_ih, *p_n_pos_in_item + n_counter) ) {
	brelse(p_s_unfm_bh);
	return 1;
      }
    }

    /* Sync UNFM. */
    if ( reiserfs_sync_block(th, p_s_unfm_bh, n_wait, NULL) )
      return -1;

    *p_n_synced += 1;

  }
  return 0;
#endif
}


/* Sync a reiserfs file item. */
static int  sync_file_item(
	      struct reiserfs_transaction_handle *th,
              struct inode  * p_s_inode,        /* Pointer to the file inode.             */
              struct path   * p_s_path_to_item, /* Pointer to the path to the found item. */
              int           * p_n_pos_in_item,  /* Position in the found item.            */
              int             n_wait,           /* Sync parameter.                        */
              unsigned long * p_n_synced  /* Returned value. Bytes number were
                                                    synced in this call.            */
            ) {

  return 0;

#if 0
  struct item_head      s_ih;
  int                   n_ret_value;

  *p_n_synced = 0;
  /* Copy found item header. */
  copy_item_head(&s_ih, PATH_PITEM_HEAD(p_s_path_to_item));

  /* Sync found item. */
  if ( (n_ret_value = reiserfs_sync_block(th, PATH_PLAST_BUFFER(p_s_path_to_item), n_wait, p_s_path_to_item)) )
    return n_ret_value;
  if ( is_direct_le_ih (&s_ih) )  {
    /* s_ih.ih_item_len bytes was synced. */
    *p_n_synced = s_ih.ih_item_len;
    return 0;
  }
  /* Sync unformatted nodes pointed by the synced indirect item. */
  n_ret_value = sync_unf_nodes(th, p_s_inode, &s_ih, p_n_pos_in_item,
			       p_s_path_to_item, n_wait, p_n_synced);

#ifdef REISERFS_CHECK
  if (!n_wait &&  ! *p_n_synced && ! n_ret_value )
    reiserfs_warning("sync_file_item: no unformatted nodes were synced (ih %h) pass %d\n", &s_ih, n_wait);
#endif

  *p_n_synced *= p_s_inode->i_sb->s_blocksize;
  return n_ret_value;
#endif
}


/* Sync a reiserfs file items. */
static int sync_file_items(
	     struct reiserfs_transaction_handle *th,
             struct inode  * p_s_inode,
	     int             n_wait
           ) {

  
  return 0;

#if 0
  struct cpu_key      s_item_key;
  INITIALIZE_PATH (s_path_to_item);
  int                 n_pos_in_item,
                      n_ret_value = 0,
		      n_inode_ret_value = 0 ;
  unsigned long       n_synced;
  unsigned long       orig_alloc ;
  struct super_block *orig_super ;


  needs changes


  /* Form key to search for the first file item. */
  make_cpu_key (&s_item_key, p_s_inode, 1,
		tail_offset (p_s_inode) == 0 ? TYPE_DIRECT : TYPE_INDIRECT, 0);

  /* While next file item is presented in the tree. */
  while ( search_for_position_by_key(p_s_inode->i_sb, &s_item_key, &s_path_to_item) == POSITION_FOUND )  {

#ifdef REISERFS_CHECK
    if ( is_direntry_le_ih(PATH_PITEM_HEAD(&s_path_to_item)) ||
	 is_statdata_le_ih(PATH_PITEM_HEAD(&s_path_to_item)) )
      reiserfs_panic (p_s_inode->i_sb, "PAP-16030: sync_file_items: unexpected item type");
#endif

    /* Synchronize the current item. */  
    if ( (n_ret_value = sync_file_item(th, p_s_inode, &s_path_to_item, &n_pos_in_item,
				       n_wait, &n_synced)) )
      break;
    /* Update key to search for the next file item. */
    set_cpu_key_k_offset (&s_item_key, cpu_key_k_offset (&s_item_key) + n_synced);
/*    if ( (k_offset (&s_item_key) += n_synced) >= p_s_inode->u.reiserfs_i.i_first_direct_byte )*/
    if ( has_tail (p_s_inode) && 
	 cpu_key_k_offset (&s_item_key) >= tail_offset (p_s_inode) + 1 )
      set_cpu_key_k_type (&s_item_key, TYPE_DIRECT);

    if (journal_transaction_should_end(th, th->t_blocks_allocated)) {
      decrement_counters_in_path(&s_path_to_item);
      if ((n_inode_ret_value = reiserfs_sync_inode(th, p_s_inode)) < 0) {
	n_ret_value = n_inode_ret_value ;
        break ;
      }
      orig_super = th->t_super ;
      orig_alloc = th->t_blocks_allocated ;
      journal_end(th, th->t_super, th->t_blocks_allocated) ;
      journal_begin(th, orig_super, orig_alloc) ;
    }
  }

  decrement_counters_in_path(&s_path_to_item);
  return n_ret_value;

#endif
}

#endif /* #if for sync code the 2.3 port does not use anymore */

/* Sync a reiserfs file. */
int reiserfs_sync_file(
      struct file   * p_s_filp,
      struct dentry * p_s_dentry
    ) {
  struct inode * p_s_inode = p_s_dentry->d_inode;
  struct reiserfs_transaction_handle th ;
  int n_err = 0;
  int windex ;
  int jbegin_count = 1 ;

  if ( S_ISDIR(p_s_inode->i_mode) )
    reiserfs_panic (p_s_inode->i_sb,
		    "PAP-16040: reiserfs_sync_file: attempt to sync directory using reiserfs_sync_file()");

  if ( ! (S_ISREG(p_s_inode->i_mode) || S_ISLNK(p_s_inode->i_mode)) )
    return -EINVAL;

  lock_kernel() ;
  /* step one, flush all dirty buffers in the file's page map to disk */
  increment_i_read_sync_counter(p_s_inode);
  n_err = generic_buffer_fdatasync(p_s_inode, 0, ~0UL) ;
  decrement_i_read_sync_counter(p_s_inode);

  /* step two, commit the current transaction to flush any metadata
  ** changes
  */
  journal_begin(&th, p_s_inode->i_sb, jbegin_count) ;
  windex = push_journal_writer("sync_file") ;
  reiserfs_update_sd(&th, p_s_inode, READ_BLOCKS);
  pop_journal_writer(windex) ;
  journal_end_sync(&th, p_s_inode->i_sb,jbegin_count) ;
  unlock_kernel() ;
  return ( n_err < 0 ) ? -EIO : 0;
}



#endif /* __KERNEL__ */

