/*
 * Copyright 1996, 1997, 1998 Hans Reiser, see reiserfs/README for licensing and copyright details
 */
#ifdef __KERNEL__

#include <linux/string.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/reiserfs_fs.h>
#include <linux/stat.h>
#include <linux/smp_lock.h>
#include <asm/uaccess.h>

#else

#include "nokernel.h"

#endif


extern struct key  MIN_KEY;

static ssize_t reiserfs_dir_read  (struct file * filp, char * buf, size_t count, loff_t * ppos)
{
  return -EISDIR;
}

/*static loff_t reiserfs_llseek(struct file *file, loff_t offset, int origin);*/
static int reiserfs_readdir (struct file *, void *, filldir_t);
int reiserfs_dir_fsync(struct file *filp, struct dentry *dentry)  ;

struct file_operations reiserfs_dir_operations = {
    read:	reiserfs_dir_read,
    readdir:	reiserfs_readdir,
    fsync:	reiserfs_dir_fsync,
};

/*
 * directories can handle most operations...
 */
struct inode_operations reiserfs_dir_inode_operations = {
  //&reiserfs_dir_operations,	/* default_file_ops */
    create:	reiserfs_create,
    lookup:	reiserfs_lookup,
    link:	reiserfs_link,
    unlink:	reiserfs_unlink,
    symlink:	reiserfs_symlink,
    mkdir:	reiserfs_mkdir,
    rmdir:	reiserfs_rmdir,
    mknod:	reiserfs_mknod,
    rename:	reiserfs_rename,
};

int reiserfs_dir_fsync(struct file *filp, struct dentry *dentry) {
  int ret = 0 ;
  int windex ;
  struct reiserfs_transaction_handle th ;

  if (!dentry || !dentry->d_inode) {
     return file_fsync(filp, dentry) ;
  }

  journal_begin(&th, dentry->d_inode->i_sb, 1) ;
  windex = push_journal_writer("dir_fsync") ;
  reiserfs_prepare_for_journal(th.t_super, SB_BUFFER_WITH_SB(th.t_super), 1) ;
  journal_mark_dirty(&th, dentry->d_inode->i_sb, SB_BUFFER_WITH_SB (dentry->d_inode->i_sb)) ;
  pop_journal_writer(windex) ;
  journal_end_sync(&th, dentry->d_inode->i_sb, 1) ;

  return ret ;
}


/* read ahead can be used also for directory */
/* void * dirent is struct getdents_callback * buf */
static int reiserfs_readdir (struct file * filp, void * dirent, filldir_t filldir)
{
  struct inode *inode = filp->f_dentry->d_inode;
  struct cpu_key pos_key;	/* key of current position in the directory (key of directory entry) */
  INITIALIZE_PATH (path_to_entry);
  struct buffer_head * bh;
  int item_num, entry_num;

  struct key * rkey;
  struct item_head * ih;
  int search_res;

  reiserfs_check_lock_depth("readdir") ;

  if (!inode || !inode->i_sb || !S_ISDIR (inode->i_mode))
    return -EBADF;

  /* form key for search the next directory entry using f_pos field of
     file structure */
  
  make_cpu_key (&pos_key, inode, (filp->f_pos) ? (filp->f_pos) : DOT_OFFSET,
		TYPE_DIRENTRY, 3);

/*  reiserfs_warning ("reiserfs_readdir 1: f_pos = %Ld\n", filp->f_pos);*/

  while (1) {
    /* search the directory item, containing entry with specified key */
    search_res = search_by_entry_key (inode->i_sb, &pos_key, &path_to_entry);
    entry_num = path_to_entry.pos_in_item;
    bh = PATH_PLAST_BUFFER (&path_to_entry);
    item_num = PATH_LAST_POSITION (&path_to_entry);
    ih = B_N_PITEM_HEAD (bh, item_num);
		
#ifdef REISERFS_CHECK
    /* we must have found item, that is item of this directory, */
    if (COMP_SHORT_KEYS (&(ih->ih_key), &pos_key))
      reiserfs_panic (inode->i_sb, "vs-9000: reiserfs_readdir: found item %h does not match to dir we readdir %k",
		      ih, &pos_key);
      
    if (item_num > B_NR_ITEMS (bh) - 1)
      reiserfs_panic (inode->i_sb, "vs-9005: reiserfs_readdir: item_num == %d, item amount == %d",
		      item_num, B_NR_ITEMS (bh));
      
    /* and entry must be not more than number of entries in the item */
    if (I_ENTRY_COUNT (ih) < entry_num)
      reiserfs_panic (inode->i_sb, "vs-9010: reiserfs_readdir: entry number is too big %d (%d)",
		      entry_num, I_ENTRY_COUNT (ih));
#endif	/* REISERFS_CHECK */

    if (search_res == POSITION_FOUND || entry_num < I_ENTRY_COUNT (ih)) {
      /* go through all entries in the directory item beginning from the entry, that has been found */
      struct reiserfs_de_head * deh = B_I_DEH (bh, ih) + entry_num;

      for (; entry_num < I_ENTRY_COUNT (ih); entry_num ++, deh ++) {
	int namelen;
	char * name;
	off_t off;

	if (!de_visible (deh))
	  /* it is hidden entry */
	  continue;
	namelen = entry_length (bh, ih, entry_num);
	name = B_I_DEH_ENTRY_FILE_NAME (bh, ih, deh);
	if (!name[namelen - 1])
	  namelen = strlen (name);
	
	off = deh_offset (deh);
	if (deh_offset (deh) != (loff_t)off) {
	  reiserfs_warning ("vs-9015: reiserfs_readdir: "
			    "dir offset set incorrect");
	  pathrelse (&path_to_entry);
	  return -EINVAL;
	}
	  
	if (filldir (dirent, name, namelen, off, DEH_OBJECTID (deh)) < 0) {
/*  reiserfs_warning ("reiserfs_readdir returned 1: f_pos = %Ld\n", filp->f_pos);*/
	  pathrelse (&path_to_entry);
	  return 0;
	}

	/* set position in directory to the next entry */
	filp->f_pos = deh_offset (deh) + 1;

      } /* for */
    }
    /* item we went through is last item of node. Using right
       delimiting key check is it directory end */
    if (item_num == B_NR_ITEMS (bh) - 1) {
      rkey = get_rkey (&path_to_entry, inode->i_sb);
      if (! comp_le_keys (rkey, &MIN_KEY)) {
#ifdef REISERFS_CHECK
	printk ("reiserfs_readdir could not calculate rigth delimiting key (path has been changed)\n");
#endif
	/* set pos_key to key, that is the smallest and greater
	   that key of the last entry in the item */
	set_cpu_key_k_offset (&pos_key, filp->f_pos);
	continue;
      }

      if ( COMP_SHORT_KEYS (rkey, &pos_key)) {
	/* end of the directory */
	pathrelse (&path_to_entry);
/*  reiserfs_warning ("reiserfs_readdir returned 2: f_pos = %Ld\n", filp->f_pos);*/
	return 0;
      }
      /* directory continues in the right neighboring block */
      copy_key (&pos_key, rkey);
      continue;
    }

    /* directory item is not a last item of the node. End of the directory is achieved */
    pathrelse (&path_to_entry);
/*  reiserfs_warning ("reiserfs_readdir returned 3: f_pos = %Ld\n", filp->f_pos);*/
    return 0;
  } /* while */

  reiserfs_check_path(&path_to_entry) ;
  return 0;
}


#if 0
static loff_t reiserfs_llseek(struct file *file, loff_t offset, int origin)
{
  long long retval;

  switch (origin) {
  case 2:
    offset += file->f_dentry->d_inode->i_size;
    break;
  case 1:
    offset += file->f_pos;
  }
  retval = -EINVAL;
 /* if (offset >= 0) {*/
    if (offset != file->f_pos) {
      file->f_pos = offset;
      file->f_reada = 0;
      file->f_version = ++event;
    }
    retval = offset;
/*  }*/
  return retval;
}
#endif




















