/*
 * Copyright 1996, 1997, 1998 Hans Reiser, see reiserfs/README for licensing and copyright details
 */
#ifdef __KERNEL__

#include <linux/sched.h>
#include <linux/reiserfs_fs.h>
#include <linux/locks.h>
#include <linux/smp_lock.h>
#include <asm/uaccess.h>

#else

#include "nokernel.h"

#endif


void reiserfs_delete_inode (struct inode * inode)
{
  int jbegin_count = JOURNAL_PER_BALANCE_CNT * 2; 
  int windex ;
  struct reiserfs_transaction_handle th ;

  
  lock_kernel() ; 

  /* The = 0 happens when we abort creating a new inode for some reason like lack of space.. */
  if (INODE_PKEY(inode)->k_objectid != 0) {
    down (&inode->i_sem); 

    journal_begin(&th, inode->i_sb, jbegin_count) ;
    windex = push_journal_writer("delete_inode") ;

    reiserfs_delete_object (&th, inode);
    reiserfs_remove_page_from_flush_list(&th, inode) ;
    pop_journal_writer(windex) ;
    reiserfs_release_objectid (&th, inode->i_ino, inode->i_sb);

    journal_end(&th, inode->i_sb, jbegin_count) ;

    up (&inode->i_sem); 
  } else {
    /* no object items are in the tree */
    ;
  }
  clear_inode (inode); /* note this must go after the journal_end to prevent deadlock */
  unlock_kernel() ;
}


#if 0
static void copy_data_blocks_to_inode (struct inode * inode, struct item_head * ih, __u32 * ind_item)
{
  int first_log_block = (ih->ih_key.k_offset - 1) / inode->i_sb->s_blocksize; /* first log block addressed by indirect item */
  int i, j;
  
  for (i = first_log_block, j = 0; i < REISERFS_N_BLOCKS && j < I_UNFM_NUM (ih); i ++, j ++) {
#ifdef REISERFS_CHECK
    if (inode->u.reiserfs_i.i_data [i] && inode->u.reiserfs_i.i_data [i] != ind_item [j])
      reiserfs_panic (inode->i_sb, "vs-13000: copy_data_blocks_to_inode: "
							 "log block %d, data block %d is seet and doe not match to unfmptr %d",
							 i, inode->u.reiserfs_i.i_data [i], ind_item [j]);
#endif
    inode->u.reiserfs_i.i_data [i] = ind_item [j];
  }
}
#endif/*0*/



void _make_cpu_key (struct cpu_key * key, int version, __u32 dirid, __u32 objectid, 
		    loff_t offset, int type)
{
  key->version = version;

  key->on_disk_key.k_dir_id = dirid;
  key->on_disk_key.k_objectid = objectid;
  set_cpu_key_k_offset (key, offset);
  set_cpu_key_k_type (key, type);  
}


/* take base of inode_key (it comes from inode always) (dirid, objectid) and version from an inode, set
   offset and type of key */
void make_cpu_key (struct cpu_key * key, const struct inode * inode, loff_t offset,
		   int type, int key_length)
{
  _make_cpu_key (key, inode_items_version (inode), le32_to_cpu (INODE_PKEY (inode)->k_dir_id),
					  le32_to_cpu (INODE_PKEY (inode)->k_objectid), offset, type);
  key->key_length = key_length;
}


//
// when key is 0, do not set version and short key
//
inline void make_le_item_head (struct item_head * ih, struct cpu_key * key, int version,
			       loff_t offset, int type, int length, int entry_count)
{
  if (key) {
	 ih->ih_key.k_dir_id = cpu_to_le32 (key->on_disk_key.k_dir_id);
	 ih->ih_key.k_objectid = cpu_to_le32 (key->on_disk_key.k_objectid);
  }
  ih->ih_version = cpu_to_le16 (version);
  set_le_ih_k_offset (ih, offset);
  set_le_ih_k_type (ih, type);
  ih->ih_item_len = cpu_to_le16 (length);
  /*    set_ih_free_space (ih, 0);*/
  // for directory items it is entry count, for directs and stat
  // datas - 0xffff, for indirects - 0
  ih->u.ih_entry_count = cpu_to_le16 (entry_count);
}


//
// FIXME: we might cache recently accessed indirect item (or at least
// first 15 pointers just like ext2 does
//
static int got_from_inode (struct inode * inode, b_blocknr_t * pblock)
{
  return 0;
}

/* people who call journal_begin with a page locked must call this
** BEFORE calling journal_begin
*/
static int prevent_flush_page_lock(struct super_block *s,
                                   struct page *page, 
				   struct inode *inode) {
  struct reiserfs_page_list *pl ;
  /* we don't care if the inode has a stale pointer from an old
  ** transaction
  */
  if(inode->u.reiserfs_i.i_conversion_trans_id != SB_JOURNAL(s)->j_trans_id) {
    return 0 ;
  }
  pl = inode->u.reiserfs_i.i_converted_page ;
  if (pl && pl->page == page) {
    pl->do_not_lock = 1 ;
  }
  return 0 ;
 
}
/* people who call journal_end with a page locked must call this
** AFTER calling journal_end
*/
static int allow_flush_page_lock(struct super_block *s,
                                   struct page *page, 
				   struct inode *inode) {

  struct reiserfs_page_list *pl ;
  /* we don't care if the inode has a stale pointer from an old
  ** transaction
  */
  if(inode->u.reiserfs_i.i_conversion_trans_id != SB_JOURNAL(s)->j_trans_id) {
    return 0 ;
  }
  pl = inode->u.reiserfs_i.i_converted_page ;
  if (pl && pl->page == page) {
    pl->do_not_lock = 0 ;
  }
  return 0 ;
 
}

#if 0

/* convert logical file block to block number if corresponding
   unformatted node. */
static b_blocknr_t reiserfs_bmap (struct inode * inode, long block)
{
  struct cpu_key offset_key;
  INITIALIZE_PATH (path_to_blocknr);
  struct buffer_head * bh;
  struct item_head * ih;
  b_blocknr_t blocknr;

  if (block < 0) {
	 reiserfs_warning ("reiserfs_bmap: block < 0\n");
	 return 0;
  }
  
  make_cpu_key (&offset_key, inode, 
					 (loff_t)block * inode->i_sb->s_blocksize + 1, TYPE_INDIRECT);

  if (byte_in_tail (inode, cpu_key_k_offset (&offset_key)))
	 /* required offset is not in an unformatted node */
	 return 0;

  if (got_from_inode (inode, &blocknr))
      /* FIXME: if we had cached appropriate indirect item here using
	 it would be better of using search_by_key */
      return blocknr;

  if (search_for_position_by_key (inode->i_sb, &offset_key, &path_to_blocknr) == POSITION_NOT_FOUND) {
	 /* file has been truncated under us? */
	 pathrelse (&path_to_blocknr);
	 return 0;
  }

  bh = get_bh (&path_to_blocknr);
  ih = get_ih (&path_to_blocknr);

  if (is_indirect_le_ih (ih)) {
	 __u32 * ind_item = (__u32 *)B_I_PITEM (bh, ih);

	 /* FIXME: here we could cache indirect item or part of it in
		 the inode to avoid search_by_key in case of subsequent
		 access to file */
	 blocknr = le32_to_cpu (ind_item [path_to_blocknr.pos_in_item]);
	 pathrelse (&path_to_blocknr);
	 inode->i_sb->u.reiserfs_sb.s_bmaps ++;
	 return blocknr;
  }

  reiserfs_warning ("vs-13030: reiserfs_bmap: found item \n%h\n is not indirect one\n", ih);
  pathrelse (&path_to_blocknr);
  return 0;
}


#endif




/* we need to allocate a block for new unformatted node.  Try to figure out
   what point in bitmap reiserfs_new_blocknrs should start from. */
static b_blocknr_t find_tag (struct buffer_head * bh, struct item_head * ih,
			     __u32 * item, int pos_in_item)
{
  if (!is_indirect_le_ih (ih))
	 /* something more complicated could be here */
	 return bh->b_blocknr;

  /* for indirect item: go to left and look for the first non-hole entry in
	  the indirect item */
  if (pos_in_item == I_UNFM_NUM (ih))
	 pos_in_item --;
  while (pos_in_item >= 0) {
	 if (item [pos_in_item])
		return item [pos_in_item];
	 pos_in_item --;
  }
  return bh->b_blocknr;
}


/* reiserfs_get_block does not need to allocate a block only if it has been
   done already or non-hole position has been found in the indirect item */
static inline int allocation_needed (int retval, b_blocknr_t allocated, 
				     struct item_head * ih,
				     __u32 * item, int pos_in_item)
{
  if (allocated)
	 return 0;
  if (retval == POSITION_FOUND && is_indirect_le_ih (ih) && item[pos_in_item])
	 return 0;
  return 1;
}

static inline int indirect_item_found (int retval, struct item_head * ih)
{
  return (retval == POSITION_FOUND) && is_indirect_le_ih (ih);
}


static inline void set_block_dev_mapped (struct buffer_head * bh, 
													  b_blocknr_t block, struct inode * inode)
{
  bh->b_dev = inode->i_dev;
  bh->b_blocknr = block;
  bh->b_state |= (1UL << BH_Mapped);
}


//
// files which were created in the earlier version can not be longer,
// than 2 gb
//
int file_capable (struct inode * inode, long block)
{
  if (inode_items_version (inode) != ITEM_VERSION_1 || block < (1 << (31 - inode->i_sb->s_blocksize_bits)))
	 return 1;
  return 0;
}

static void check_gb_alloc(int gb_alloc, int gb_freed, char * caller) {
  if (gb_alloc != gb_freed) {
	 printk("check_gb_alloc %s, %d alloc, %d freed\n", caller, gb_alloc, gb_freed) ;
  }
}

static void restart_transaction(struct reiserfs_transaction_handle *th,
				struct inode *inode, struct path *path) {
  struct super_block *s = th->t_super ;
  int len = th->t_blocks_allocated ;

  pathrelse(path) ;
  reiserfs_update_sd(th, inode, READ_BLOCKS) ;
  journal_end(th, s, len) ;
  journal_begin(th, s, len) ;
}

// it is called by get_block when create == 0. Returns block number
// for 'block'-th logical block of file. When it hits direct item it
// returns 0 (being called from bmap) or read direct item into piece
// of page (bh_result)
static void _get_block_create_0 (struct inode * inode, long block,
				 struct buffer_head * bh_result)
{
    INITIALIZE_PATH (path);
    struct cpu_key key;
    struct buffer_head * bh;
    struct item_head * ih;
    int blocknr;
    char * p;
    int chars;

 repeat:

    // prepare the key to look for the 'block'-th block of file
    make_cpu_key (&key, inode,
		  (loff_t)block * inode->i_sb->s_blocksize + 1, TYPE_ANY, 3);

    if (got_from_inode (inode, &bh_result->b_blocknr)) {
	bh_result->b_dev = inode->i_dev;
	//bh_result->b_blocknr = block;
	bh_result->b_state |= (1UL << BH_Mapped);
	return;
    }

    if (search_for_position_by_key (inode->i_sb, &key, &path) != POSITION_FOUND) {
	pathrelse (&path);
	return;
    }
    
    //
    bh = get_bh (&path);
    ih = get_ih (&path);
    if (is_indirect_le_ih (ih)) {
	__u32 * ind_item = (__u32 *)B_I_PITEM (bh, ih);
	
	/* FIXME: here we could cache indirect item or part of it in
	   the inode to avoid search_by_key in case of subsequent
	   access to file */
	blocknr = le32_to_cpu (ind_item [path.pos_in_item]);
	if (blocknr) {
	    bh_result->b_dev = inode->i_dev;
	    bh_result->b_blocknr = blocknr;
	    bh_result->b_state |= (1UL << BH_Mapped);
	}
	pathrelse (&path);
	return;
    }

    if (bh_result->b_dev != inode->i_dev)
	// FIXME: generic_block_map called that. We trust to
	// create_empty_buffers who should have set b_dev for all page
	// buffers. As it is not done for bh_result, we are called by
	// generic_block_map and therefore there is no need to read
	// contents of direct item. Note also: lilo fails here
	return;

    if (conversion_in_progress (inode))
	goto repeat;

    lock_tail (inode);

    // read file tail into part of page
    p = bh_result->b_data;
    do {
	if (!is_direct_le_ih (ih))
	    BUG ();
	chars = le16_to_cpu (ih->ih_item_len) - path.pos_in_item;
	memcpy (p, B_I_PITEM (bh, ih) + path.pos_in_item, chars);
	p += chars;

	if (PATH_LAST_POSITION (&path) != (B_NR_ITEMS (bh) - 1))
	    // we done, if read direct item is not the last item of
	    // node FIXME: we could try to check right delimiting key
	    // to see whether direct item continues in the right
	    // neighbor or rely on i_size
	    break;

	// update key to look for the next piece
	set_cpu_key_k_offset (&key, cpu_key_k_offset (&key) + chars);
	if (search_for_position_by_key (inode->i_sb, &key, &path) != POSITION_FOUND)
	    break;
	bh = get_bh (&path);
	ih = get_ih (&path);
    } while (1);

    unlock_tail (inode);
    pathrelse (&path);
    
    // FIXME: b_blocknr == 0 here. but b_data contains correct data
    // from tail. ll_rw_block will skip uptodate buffers
    mark_buffer_uptodate (bh_result, 1);
    bh_result->b_state |= (1UL << BH_Mapped);

    return;
}


int reiserfs_get_block (struct inode * inode, long block,
			struct buffer_head * bh_result, int create)
{
    int repeat, retval;
    unsigned long tag;
    b_blocknr_t phys;
    INITIALIZE_PATH(path);
    int pos_in_item;
    struct cpu_key key;
    loff_t offset;
    struct buffer_head * bh, * unbh = 0;
    struct item_head * ih, tmp_ih;
    __u32 * item;
    int done;
    int fs_gen;
    int windex ;
    struct reiserfs_transaction_handle th ;
    int jbegin_count = JOURNAL_PER_BALANCE_CNT * 3 ;
    int version;
    int gb_freed = 0 ;
    int gb_alloc = 0 ;

    lock_kernel() ;
    version = inode_items_version (inode);

    if (!file_capable (inode, block)) {
	unlock_kernel() ;
	return -EFBIG;
    }

    /* if !create, we aren't changing the FS, so we don't need to
    ** log anything, so we don't need to start a transaction
    */
    if (!create) {
	/* find number of block-th logical block of the file */
	_get_block_create_0 (inode, block, bh_result);
	unlock_kernel() ;
#ifdef REISERFS_CHECK
	if (bh_result->b_end_io == reiserfs_end_buffer_io_sync) {
	    printk("1warning, get_block returning buffer with the wrong end_io func\n") ;
	}
#endif
	return 0;
    }

    if (block < 0) {
	unlock_kernel();
	return -EIO;
    }

    prevent_flush_page_lock(inode->i_sb, bh_result->b_page, inode) ;
    journal_begin(&th, inode->i_sb, jbegin_count) ;
    windex = push_journal_writer("reiserfs_get_block") ;
  
    /* create block-th block in the file */
    phys = 0;
  
    /* set the key of the first byte in the 'block'-th block of file */
    offset = (loff_t)block * inode->i_sb->s_blocksize + 1;
    // note, we are sure, that there is no conversion in the progress
    if (conversion_in_progress (inode))
	BUG ();
    make_cpu_key (&key, inode, offset, TYPE_ANY, 3/*key length*/);

 research:


    retval = search_for_position_by_key (inode->i_sb, &key, &path);
    bh = get_bh (&path);
    ih = get_ih (&path);
    item = get_item (&path);
    pos_in_item = path.pos_in_item;

    fs_gen = get_generation (inode->i_sb);

    if (allocation_needed (retval, phys, ih, item, pos_in_item)) {
	/* struct buffer_head * tmp; */

	/* we have to allocate block for the unformatted node */
	copy_item_head (&tmp_ih, ih);
	tag = find_tag (bh, ih, item, pos_in_item);
	repeat = reiserfs_new_unf_blocknrs (&th, inode->i_sb, &phys, tag, 1, 0/* for unformatted node, do not use spare space */);
	if (repeat == NO_DISK_SPACE) {
	    /* restart the transaction to give the journal a chance to free
	    ** some blocks.  releases the path, so we have to go back to
	    ** research if we succeed on the second try
	    */
	    restart_transaction(&th, inode, &path) ; 
	    repeat = reiserfs_new_unf_blocknrs (&th, inode->i_sb, &phys, tag, 
	                                        1, 0);
	    if (repeat != NO_DISK_SPACE) {
	      phys = 0 ; /* just in case it got changed somehow */
	      goto research ;
	    }
	    reiserfs_update_sd(&th, inode, READ_BLOCKS) ;
	    pop_journal_writer(windex) ;
	    journal_end(&th, inode->i_sb, jbegin_count) ;
	    allow_flush_page_lock(inode->i_sb, bh_result->b_page, inode) ;
	    unlock_kernel() ;
	    check_gb_alloc(gb_alloc, gb_freed, "1") ;
	    return -ENOSPC;
	}

	if (fs_changed (fs_gen, inode->i_sb) && item_moved (&tmp_ih, &path)) {
	    goto research;
	}
    }

    if (indirect_item_found (retval, ih)) {
	/* 'block'-th block is in the file already (there is
	   corresponding cell in some indirect item). But it may be
	   zero unformatted node pointer (hole) */
	if (!item[pos_in_item]) {
	    /* use allocated block to plug the hole */
	    reiserfs_prepare_for_journal(inode->i_sb, bh, 1) ;
	    if (fs_changed (fs_gen, inode->i_sb) && item_moved (&tmp_ih, &path)) {
		reiserfs_restore_prepared_buffer(inode->i_sb, bh) ;
		goto research;
	    }
	    bh_result->b_state |= (1UL << BH_New);
	    item[pos_in_item] = cpu_to_le32 (phys);
	    journal_mark_dirty (&th, inode->i_sb, bh);
	}
	set_block_dev_mapped (bh_result, le32_to_cpu (item[pos_in_item]), inode);
	pathrelse (&path);
	pop_journal_writer(windex) ;
	journal_end(&th, inode->i_sb, jbegin_count) ;
	allow_flush_page_lock(inode->i_sb, bh_result->b_page, inode) ;
	unlock_kernel() ;
	if (bh_result->b_end_io == reiserfs_end_buffer_io_sync) {
	    printk("2warning, get_block returning buffer with the wrong end_io func\n") ;
	}
	check_gb_alloc(gb_alloc, gb_freed, "2") ;
	 
	/* the item was found, so the file was not grown or changed. 
	** there is no need to make sure the inode is updated with this 
	** transaction
	*/
	return 0;
    }

    /* desired position is not found or is in the direct item. We have
       to append file with holes up to 'block'-th block converting
       direct items to indirect one if necessary */
    done = 0;
    do {
	if (is_statdata_le_ih (ih)) {
	    __u32 unp = 0;

	    /* indirect item has to be inserted */
	    make_le_item_head (&tmp_ih, &key, version, 1, TYPE_INDIRECT, 
			       UNFM_P_SIZE, 0/* free_space */);

	    if (cpu_key_k_offset (&key) == 1) {
		/* we are going to add 'block'-th block to the file. Use
		   allocated block for that */
		unp = cpu_to_le32 (phys);
		set_block_dev_mapped (bh_result, phys, inode);
		bh_result->b_state |= (1UL << BH_New);
		done = 1;
	    }
	    PATH_LAST_POSITION(&path) ++;
	    if (reiserfs_insert_item (&th, inode->i_sb, &path, &tmp_ih, 
				      (char *)&unp) != 0) {
		reiserfs_free_block (&th, inode->i_sb, phys);
		/*up (&inode->i_sem);*/
		reiserfs_update_sd (&th, inode, READ_BLOCKS) ;
		pop_journal_writer(windex) ;
		journal_end(&th, inode->i_sb, jbegin_count) ;
		allow_flush_page_lock(inode->i_sb, bh_result->b_page, inode) ;
		unlock_kernel() ;
		check_gb_alloc(gb_alloc, gb_freed, "3") ;
		reiserfs_check_path(&path) ;
		return -ENOSPC;
	    }
	    if (unp)
		inode->i_blocks += inode->i_sb->s_blocksize / 512;
	    //mark_tail_converted (inode);
	} else if (is_direct_le_ih (ih)) {
	    /* direct item has to be converted */
	    loff_t tail_offset;

	    tail_offset = ((le_ih_k_offset (ih) - 1) & ~(inode->i_sb->s_blocksize - 1)) + 1;
	    if (tail_offset == cpu_key_k_offset (&key)) {
		/* direct item we just found fits into block we have
                   to map. Convert it into unformatted node: use
                   bh_result for the convertion */
		set_block_dev_mapped (bh_result, phys, inode);
		unbh = bh_result;
		done = 1;
	    } else {
		/* we have to padd file tail stored in direct item(s)
		   up to block size and convert it to unformatted
		   node. FIXME: this should also get into page cache */
		if (!unbh) {
		    copy_item_head (&tmp_ih, ih);
			 
		    fs_gen = get_generation (inode->i_sb);

		    get_new_buffer (&th, inode->i_sb, bh, &unbh, &path);
		    if (!unbh) {
			/* restart_transaction calls pathrelse for us */
			restart_transaction(&th, inode, &path) ;
			get_new_buffer (&th, inode->i_sb, bh, &unbh, &path);
			if (unbh) {
			  goto search_again ;
			}
			reiserfs_free_block (&th, inode->i_sb, phys);
			reiserfs_update_sd(&th, inode, READ_BLOCKS) ;
			pop_journal_writer(windex) ;
			journal_end(&th, inode->i_sb, jbegin_count) ;
			allow_flush_page_lock(inode->i_sb, bh_result->b_page, 
			                      inode) ;
			unlock_kernel() ;
			check_gb_alloc(gb_alloc, gb_freed, "4") ;
			return -ENOSPC;
		    }
		    gb_alloc++ ;
		    //unbh->b_state |= (1UL << BH_New);
		    if (fs_changed (fs_gen, inode->i_sb) && 
		        item_moved (&tmp_ih, &path)) {
			goto search_again ;
		    }
		}
	    }
	    mark_buffer_uptodate (unbh, 1);
	    if (direct2indirect (&th, inode, &path, unbh, tail_offset) != 0) {
		if (!done) {
		    // free what has been allocated by get_new_buffer
		    unsigned long tmp = unbh->b_blocknr;

		    bforget (unbh);
		    reiserfs_free_block (&th, inode->i_sb, tmp);
		}
		reiserfs_free_block (&th, inode->i_sb, phys);
		reiserfs_update_sd(&th, inode, READ_BLOCKS) ;
		pop_journal_writer(windex) ;
		journal_end(&th, inode->i_sb, jbegin_count) ;
		allow_flush_page_lock(inode->i_sb, bh_result->b_page, inode) ;
		unlock_kernel() ;
		check_gb_alloc(gb_alloc, gb_freed, "4") ;
		reiserfs_check_path(&path) ;
		return -ENOSPC;
	    }
	    /* we've converted the tail, so we must 
	    ** flush unbh before the transaction commits
	    */
	    reiserfs_add_page_to_flush_list(&th, inode, unbh) ;
		  
	    if (!done) {
		// unbh was acquired by get_new_buffer
		brelse (unbh);
		gb_freed++ ;
	    }
	    inode->i_blocks += inode->i_sb->s_blocksize / 512;
	    //mark_tail_converted (inode);
	} else {
	    /* append indirect item with holes if needed, when appending
	       pointer to 'block'-th block use block, which is already
	       allocated */
	    struct cpu_key tmp_key;
	    struct unfm_nodeinfo un = {0, 0};

	    /* indirect item has to be appended, set up key of that position */
	    make_cpu_key (&tmp_key, inode,
			  le_key_k_offset (version, &(ih->ih_key)) + le_ih_bytes_number (ih, bh->b_size),
			  TYPE_INDIRECT, 3);// key type is unimportant
		  
	    if (cpu_key_k_offset (&tmp_key) == cpu_key_k_offset (&key)) {
		/* we are going to add target block to the file. Use allocated
		   block for that */
		un.unfm_nodenum = cpu_to_le32 (phys);
		set_block_dev_mapped (bh_result, phys, inode);
		bh_result->b_state |= (1UL << BH_New);
		done = 1;
	    } else {
		/* paste hole to the indirect item */
	    }
	    if (pos_in_item != ih->ih_item_len / UNFM_P_SIZE)
		reiserfs_panic (inode->i_sb, "vs-: reiserfs_get_block: "
				"invalid position for append");
	    if (reiserfs_paste_into_item (&th, inode->i_sb, &path,
					  &tmp_key, (char *)&un, UNFM_P_SIZE) != UNFM_P_SIZE) {
		reiserfs_free_block (&th, inode->i_sb, phys);
		reiserfs_update_sd(&th, inode, READ_BLOCKS) ;
		pop_journal_writer(windex) ;
		journal_end(&th, inode->i_sb, jbegin_count) ;
		allow_flush_page_lock(inode->i_sb, bh_result->b_page, inode) ;
		unlock_kernel() ;
		check_gb_alloc(gb_alloc, gb_freed, "5") ;
		reiserfs_check_path(&path) ;
		return -ENOSPC;
	    }
	    if (un.unfm_nodenum)
		inode->i_blocks += inode->i_sb->s_blocksize / 512;
	    //mark_tail_converted (inode);
	}
		
	if (done == 1)
	    break;
	 
	/* this loop could log more blocks than we had originally asked
	** for.  So, we have to allow the transaction to end if it is
	** too big or too full.  Update the inode so things are 
	** consistent if we crash before the function returns
	**
	** release the path so that anybody waiting on the path before
	** ending their transaction will be able to continue.
	*/
	if (journal_transaction_should_end(&th, th.t_blocks_allocated)) {
	  int orig_len_alloc = th.t_blocks_allocated ;
	  pathrelse (&path);
	  reiserfs_update_sd(&th, inode, READ_BLOCKS) ;
	  journal_end(&th, inode->i_sb, orig_len_alloc) ;
	  journal_begin(&th, inode->i_sb, orig_len_alloc) ;
	}
search_again:
	retval = search_for_position_by_key (inode->i_sb, &key, &path);
	if (retval == POSITION_FOUND)
	    reiserfs_panic (inode->i_sb, "vs-: reiserfs_get_block: "
			    "%k should not be found", &key);
	bh = get_bh (&path);
	ih = get_ih (&path);
	item = get_item (&path);
	pos_in_item = path.pos_in_item;
    } while (1);
	 
    pathrelse (&path);
    reiserfs_update_sd(&th, inode, READ_BLOCKS) ;
    pop_journal_writer(windex) ;
    journal_end(&th, inode->i_sb, jbegin_count) ;
    allow_flush_page_lock(inode->i_sb, bh_result->b_page, inode) ;
    unlock_kernel() ;
    if (bh_result->b_end_io == reiserfs_end_buffer_io_sync) {
	printk("3warning, get_block returning buffer with the wrong end_io func\n") ;
    }
    check_gb_alloc(gb_alloc, gb_freed, "6") ;
    return 0;
}


//
// BAD: new directories have stat data of new type and all other items
// of old type. Version stored in the inode says about body items, so
// in update_stat_data we can not rely on inode, but have to check
// item version directly
//


static void init_inode (struct inode * inode, struct path * path)
{
  struct buffer_head * bh;
  struct item_head * ih;
  __u32 rdev;
  int version;

  bh = PATH_PLAST_BUFFER (path);
  ih = PATH_PITEM_HEAD (path);


  copy_key (INODE_PKEY (inode), &(ih->ih_key));

  if (stat_data_v1 (ih)) {
	 struct stat_data_v1 * sd = (struct stat_data_v1 *)B_I_PITEM (bh, ih);

	 version = ITEM_VERSION_1;
	 inode->i_mode = le16_to_cpu (sd->sd_mode);
	 inode->i_nlink = le16_to_cpu (sd->sd_nlink);
	 inode->i_uid = le16_to_cpu (sd->sd_uid);
	 inode->i_gid = le16_to_cpu (sd->sd_gid);
	 inode->i_size = le32_to_cpu (sd->sd_size);
	 inode->i_atime = le32_to_cpu (sd->sd_atime);
	 inode->i_mtime = le32_to_cpu (sd->sd_mtime);
	 inode->i_mtime = le32_to_cpu (sd->sd_ctime);
	 inode->i_blocks = (inode->i_size + 511) >> 9;
	 rdev = le32_to_cpu (sd->sd_rdev);
	 //inode->u.reiserfs_i.i_first_direct_byte = le32_to_cpu (sd->sd_first_direct_byte);
  } else {
	 struct stat_data * sd = (struct stat_data *)B_I_PITEM (bh, ih);

	 /* both old and new directories have old keys */
	 version = (S_ISDIR (sd->sd_mode) ? ITEM_VERSION_1 : ITEM_VERSION_2);
	 inode->i_mode = le16_to_cpu (sd->sd_mode);
	 inode->i_nlink = le16_to_cpu (sd->sd_nlink);
	 inode->i_uid = le32_to_cpu (sd->sd_uid);
	 inode->i_size = le64_to_cpu (sd->sd_size);
	 inode->i_gid = le32_to_cpu (sd->sd_gid);
	 inode->i_mtime = le32_to_cpu (sd->sd_mtime);
	 inode->i_atime = le32_to_cpu (sd->sd_atime);
	 inode->i_ctime = le32_to_cpu (sd->sd_ctime);
	 inode->i_blksize = inode->i_sb->s_blocksize;
	 inode->i_blocks = le32_to_cpu (sd->sd_blocks);
	 rdev = le32_to_cpu (sd->u.sd_rdev);
	 //	 inode->u.reiserfs_i.i_first_direct_byte = le32_to_cpu (sd->u.sd_first_direct_byte);
  }
  //if (inode->u.reiserfs_i.i_first_direct_byte == U32_MAX)
  // inode->u.reiserfs_i.i_has_tail = 0;
  //else
  // inode->u.reiserfs_i.i_has_tail = 1;
  //  inode->u.reiserfs_i.i_has_tail =  (inode->u.reiserfs_i.i_first_direct_byte == U32_MAX) ? 0 : 1;


  pathrelse (path);
  if (S_ISREG (inode->i_mode)) {
	 inode->i_op = &reiserfs_file_inode_operations;
	 inode->i_fop = &reiserfs_file_operations;
	 inode->i_mapping->a_ops = &reiserfs_address_space_operations ;
  } else if (S_ISDIR (inode->i_mode)) {
	 inode->i_op = &reiserfs_dir_inode_operations;
	 inode->i_fop = &reiserfs_dir_operations;
  } else if (S_ISLNK (inode->i_mode)) {
	 inode->i_op = &reiserfs_symlink_inode_operations;
  } else {
	 init_special_inode(inode, inode->i_mode, rdev) ;
  }

  inode_items_version (inode) = version;

}


// update new stat data with inode fields
static void _update_stat_data (struct stat_data * sd, struct inode * inode)
{
  sd->sd_mode = cpu_to_le16 (inode->i_mode);
  sd->sd_nlink = cpu_to_le16 (inode->i_nlink);
  sd->sd_uid = cpu_to_le32 (inode->i_uid);
  sd->sd_size = cpu_to_le64 (inode->i_size);
  sd->sd_gid = cpu_to_le32 (inode->i_gid);
  sd->sd_mtime = cpu_to_le32 (inode->i_mtime);
  sd->sd_atime = cpu_to_le32 (inode->i_atime);
  sd->sd_ctime = cpu_to_le32 (inode->i_ctime);
  sd->sd_blocks = cpu_to_le32 (inode->i_blocks);
  if (S_ISCHR(inode->i_mode) || S_ISBLK(inode->i_mode))
         sd->u.sd_rdev = cpu_to_le32 (inode->i_rdev);
  else
      ;//sd->u.sd_first_direct_byte = cpu_to_le32 (inode->u.reiserfs_i.i_first_direct_byte);
}


/* NOTE, you must prepare the buffer head before sending it here,
** and then log it after the call
*/
static void update_stat_data (struct path * path, struct inode * inode)
{
  struct buffer_head * bh;
  struct item_head * ih;
  
  bh = PATH_PLAST_BUFFER (path);
  ih = PATH_PITEM_HEAD (path);

  if (!is_statdata_le_ih (ih))
	 reiserfs_panic (inode->i_sb, "vs-13065: update_stat_data: key %k, found item %h",
						  INODE_PKEY (inode), ih);
  
  if (stat_data_v1 (ih)) {
	 struct stat_data_v1 * sd;

	 sd = (struct stat_data_v1 *)B_I_PITEM (bh, ih);
	 sd->sd_mode = cpu_to_le16 (inode->i_mode);
	 sd->sd_uid = cpu_to_le16 (inode->i_uid);
	 sd->sd_gid = cpu_to_le16 (inode->i_gid);
	 sd->sd_nlink = cpu_to_le16 (inode->i_nlink);
	 sd->sd_size = cpu_to_le64 (inode->i_size);
	 sd->sd_atime = cpu_to_le32 (inode->i_atime);
	 sd->sd_ctime = cpu_to_le32 (inode->i_ctime);
	 sd->sd_mtime = cpu_to_le32 (inode->i_mtime);
	 if (S_ISCHR(inode->i_mode) || S_ISBLK(inode->i_mode))
		sd->sd_rdev = cpu_to_le32 (inode->i_rdev);
	 //	 else
	 //sd->sd_first_direct_byte = cpu_to_le32 (inode->u.reiserfs_i.i_first_direct_byte);
  } else {
	 struct stat_data * sd;
	 
         sd = (struct stat_data *)B_I_PITEM (bh, ih);
	 _update_stat_data (sd, inode);
  }

  return;
}


// FIXME: read_blocks is always READ_BLOCKS
void reiserfs_update_sd (struct reiserfs_transaction_handle *th, 
			 struct inode * inode, 
			 int read_blocks)
{
  struct cpu_key key;
  INITIALIZE_PATH(path);
  struct buffer_head *bh ;
  int fs_gen ;
  struct item_head *ih, tmp_ih ;
   

  make_cpu_key (&key, inode, SD_OFFSET, TYPE_STAT_DATA, 3);//key type is unimportant

  for(;;) {
	 /* look for the object's stat data */
	 if (search_by_key (inode->i_sb, &key, &path, DISK_LEAF_NODE_LEVEL, read_blocks) == ITEM_NOT_FOUND) {
		pathrelse(&path) ;
		if (read_blocks == DONT_READ_BLOCKS) {
		  /* this is called from if_in_ram_update_sd */
		  /*printk ("reiserfs: stat data not found in memory\n");*/
		  return;
		}
		if (inode->i_nlink == 0) {
		  /*printk ("vs-13050: reiserfs_update_sd: i_nlink == 0, stat data not found\n");*/
		  return;
		}
		reiserfs_warning ("vs-13060: reiserfs_update_sd: "
								"stat data of object %k (nlink == %d) not found (pos %d)\n", 
								INODE_PKEY (inode), inode->i_nlink);
		reiserfs_check_path(&path) ;
		return;
	 }

	 /* sigh, prepare_for_journal might schedule.  When it schedules the
	 ** FS might change.  We have to detect that, and loop back to the
	 ** search if the stat data item has moved
	 */
	 bh = get_bh(&path) ;
	 ih = get_ih(&path) ;
	 copy_item_head (&tmp_ih, ih);
	 fs_gen = get_generation (inode->i_sb);
	 reiserfs_prepare_for_journal(inode->i_sb, bh, 1) ;
	 if (fs_changed (fs_gen, inode->i_sb) && item_moved(&tmp_ih, &path)) {
		reiserfs_restore_prepared_buffer(inode->i_sb, bh) ;
		continue ;	/* Stat_data item has been moved after scheduling. */
	 }
	 break;
  }
  update_stat_data (&path, inode);
  journal_mark_dirty(th, th->t_super, bh) ; 
  pathrelse (&path);
  return;
}

void reiserfs_read_inode(struct inode *inode) {
  make_bad_inode(inode) ;
}

/* looks for stat data in the tree, and fills up the fields of in-core
   inode stat data fields */
void reiserfs_read_inode2 (struct inode * inode, void *p)
{
    INITIALIZE_PATH (path_to_sd);
    struct cpu_key key;
    unsigned long dirino;

    inode->i_op = NULL;
    inode->i_mode = 0;

    dirino = (unsigned long)p ;

    /* set version 1, version 2 could be used too, because stat data
       key is the same in both versions */
    key.version = ITEM_VERSION_1;
    key.on_disk_key.k_dir_id = dirino;
    key.on_disk_key.k_objectid = inode->i_ino;
    key.on_disk_key.u.k_offset_v1.k_offset = SD_OFFSET;
    key.on_disk_key.u.k_offset_v1.k_uniqueness = SD_UNIQUENESS;

    /* look for the object's stat data */
    if (search_item (inode->i_sb, &key, &path_to_sd) != ITEM_FOUND) {
	reiserfs_warning ("vs-13042: reiserfs_read_inode: %K not found", &key);
	pathrelse (&path_to_sd);
	return;
    }

    init_inode (inode, &path_to_sd);
    reiserfs_check_path(&path_to_sd) ; /* init inode should be relsing */

}


struct inode * reiserfs_iget (struct super_block * s, struct cpu_key * key)
{
    struct inode * inode;

    inode = iget4 (s, key->on_disk_key.k_objectid, 0, (void *)(key->on_disk_key.k_dir_id));
    if (!inode->i_op || comp_short_keys (INODE_PKEY (inode), key)) {
	reiserfs_warning ("vs-13048: reiserfs_iget: key in inode %k and key in entry %k do not match\n",
			  INODE_PKEY (inode), key);
	iput (inode);
	inode = 0;
    }
    return inode;
}




/* looks for stat data, then copies fields to it, marks the buffer
   containing stat data as dirty */
void reiserfs_write_inode (struct inode * inode)
{
  int windex ;
  struct reiserfs_transaction_handle th ;
  int jbegin_count = JOURNAL_PER_BALANCE_CNT * 3; 

  lock_kernel() ;
  journal_begin(&th, inode->i_sb, jbegin_count) ;
  windex = push_journal_writer("write_inode") ;
  reiserfs_update_sd (&th, inode, READ_BLOCKS);
  pop_journal_writer(windex) ;
  journal_end(&th, inode->i_sb, jbegin_count) ;
  unlock_kernel() ;
}


/* FIXME: no need any more. right? */
int reiserfs_sync_inode (struct reiserfs_transaction_handle *th, struct inode * inode)
{
  int err = 0;

  reiserfs_update_sd (th, inode, READ_BLOCKS);
  return err;
}


/* stat data of new object is inserted already, this inserts the item
   containing "." and ".." entries */
static int reiserfs_new_directory (struct reiserfs_transaction_handle *th, 
				 struct super_block * sb, struct item_head * ih, struct path * path, const struct inode * dir)
{
  char empty_dir [EMPTY_DIR_SIZE];
  char * body = empty_dir;
  struct cpu_key key;
    
  _make_cpu_key (&key, ITEM_VERSION_1, le32_to_cpu (ih->ih_key.k_dir_id),
					  le32_to_cpu (ih->ih_key.k_objectid), DOT_OFFSET, TYPE_DIRENTRY);

  /* compose item head for new item. Directories consist of items of
	  old type (ITEM_VEREION_1). Do not set key (second arg is 0), it is
	  done by reiserfs_new_inode */
  make_le_item_head (ih, 0, ITEM_VERSION_1, DOT_OFFSET, TYPE_DIRENTRY, EMPTY_DIR_SIZE, 2);

  make_empty_dir_item (body, ih->ih_key.k_dir_id, ih->ih_key.k_objectid,
							  le32_to_cpu (INODE_PKEY (dir)->k_dir_id), 
							  le32_to_cpu (INODE_PKEY (dir)->k_objectid));

  /* look for place in the tree for new item */
  if (search_item (sb, &key, path) == ITEM_FOUND)
	 reiserfs_panic (sb, "vs-13070: reiserfs_new_directory: "
						  "object with this key exists (%k)", &(ih->ih_key));

  /* insert item, that is empty directory item */
  return reiserfs_insert_item (th, sb, path, ih, body);
}


/* stat data of object has been inserted, this inserts the item
   containing the body of symlink */
static int reiserfs_new_symlink (struct reiserfs_transaction_handle *th, 
											struct super_block * sb, struct item_head * ih,
											struct path * path, const char * symname, int item_len)
{
  struct cpu_key key;

  _make_cpu_key (&key, ITEM_VERSION_2, 
					  le32_to_cpu (ih->ih_key.k_dir_id), 
					  le32_to_cpu (ih->ih_key.k_objectid),
					  1, TYPE_DIRECT);

  /* item head of symlink body. All new objects get new format but
	  directories */
  make_le_item_head (ih, 0, ITEM_VERSION_2, 1, TYPE_DIRECT, item_len, 0);

  /* look for place in the tree for new item */
  if (search_item (sb, &key, path) == ITEM_FOUND)
	 reiserfs_panic (sb, "vs-13080: reiserfs_new_symlink: "
						  "object with this key exists (%k)", &(ih->ih_key));

  /* insert item, that is body of symlink */
  return reiserfs_insert_item (th, sb, path, ih, symname);
}


/* inserts the stat data into the tree, and then calls
   reiserfs_new_directory (to insert ".", ".." item if new object is
   directory) or reiserfs_new_symlink (to insert symlink body if new
   object is symlink) or nothing (if new object is regular file) */
struct inode *	reiserfs_new_inode (struct reiserfs_transaction_handle *th,
											  const struct inode * dir, int mode, 
											  const char * symname, int item_len,
											  struct dentry *dentry, struct inode *inode, int * err)
{
  struct super_block * sb;
  INITIALIZE_PATH (path_to_key);
  struct cpu_key key;
  struct item_head ih;
  struct stat_data sd;
  int retvalue;
  
  if (!dir || !dir->i_nlink) {
	 *err = -EPERM;
	 return NULL;
  }

  sb = dir->i_sb;
  inode->i_sb = sb;
  inode->i_flags = inode->i_sb->s_flags;


  /* item head of new item */
  ih.ih_key.k_dir_id = INODE_PKEY (dir)->k_objectid;
  ih.ih_key.k_objectid = cpu_to_le32 (reiserfs_get_unused_objectid (th, dir->i_sb));
  if (!ih.ih_key.k_objectid) {
	 iput(inode) ;
	 *err = -ENOMEM;
	 return NULL;
  }
  ih.ih_version = cpu_to_le16 (ITEM_VERSION_2);

  make_le_item_head (&ih, 0, le16_to_cpu (ih.ih_version),
							SD_OFFSET, TYPE_STAT_DATA, SD_SIZE, MAX_US_INT/*free_space*/);

  /* key to search for correct place for new stat data */
  _make_cpu_key (&key, ITEM_VERSION_2, le32_to_cpu (ih.ih_key.k_dir_id),
					  le32_to_cpu (ih.ih_key.k_objectid), SD_OFFSET, TYPE_STAT_DATA);

  /* find proper place for inserting of stat data */
  if (search_item (sb, &key, &path_to_key) == ITEM_FOUND) {
	 pathrelse (&path_to_key);
	 iput (inode);
	 *err = -EEXIST;
	 return NULL;
  }

  /* fill stat data */
  inode->i_mode = mode;
  inode->i_nlink = (mode & S_IFDIR) ? 2 : 1;
  inode->i_uid = current->fsuid;
  inode->i_gid = (dir->i_mode & S_ISGID) ? dir->i_gid : current->fsgid;
  inode->i_mtime = inode->i_atime = inode->i_ctime = CURRENT_TIME;
  inode->i_blocks = 0;
  inode->i_size = (S_ISDIR (mode) ? EMPTY_DIR_SIZE : 
						 (S_ISLNK (mode) ? item_len : 0));
  //inode->u.reiserfs_i.i_first_direct_byte = S_ISLNK(mode) ? 1 : NO_BYTES_IN_DIRECT_ITEM;
  
  _update_stat_data (&sd, inode);
	 
  inode->i_ino = ih.ih_key.k_objectid;
  inode->i_blksize = PAGE_SIZE;
  inode->i_dev = sb->s_dev;
  inode->i_op = NULL;
  inode->i_blksize = sb->s_blocksize;
  
  // store in he in-core inode the key of stat data and version all
  // object items will have (directory items will have old offset
  // format, other new objects will consist of new items)
  memcpy (INODE_PKEY (inode), &(ih.ih_key), KEY_SIZE);
  inode_items_version (inode) = ((mode & S_IFDIR) ? ITEM_VERSION_1 : ITEM_VERSION_2);
  //  mark_file_without_tail (inode);

  /* insert the stat data into the tree */
  retvalue = reiserfs_insert_item (th, sb, &path_to_key, &ih, (char *)(&sd));
  if (retvalue == NO_DISK_SPACE) {
	 iput (inode);
	 *err = -ENOSPC;
	 reiserfs_check_path(&path_to_key) ;
	 return NULL;
  }

  if (S_ISDIR(mode)) {
	 /* insert item with "." and ".." */
	 retvalue = reiserfs_new_directory (th, sb, &ih, &path_to_key, dir);
  }

  if (S_ISLNK(mode)) {
	 /* insert body of symlink */
	 retvalue = reiserfs_new_symlink (th, sb, &ih, &path_to_key, symname, item_len);
	 //if (retvalue == 0)
	 //mark_file_with_tail (inode, 1);
  }
  if (retvalue == NO_DISK_SPACE) {
	 iput (inode);
	 *err = -ENOSPC;
	 reiserfs_check_path(&path_to_key) ;
	 return NULL;
  }

  insert_inode_hash (inode);
  // we do not mark inode dirty: on disk content matches to the
  // in-core one
  reiserfs_check_path(&path_to_key) ;
  return inode;
}

#if 0

/* we are reading into the page cache, not into any process's virtual
   memory, therefore no need for local buffer */
static int read_file_tail (struct inode * inode, loff_t offset, char * buf, int size)
{
  struct cpu_key key;
  int pos_in_item;
  INITIALIZE_PATH (path);
  struct item_head * ih;
  char * p;
  int chars, left, read;
  struct buffer_head * bh;
  int block;
  struct buffer_head * pbh[PAGE_SIZE / 512] = {0,};
  int requested[PAGE_SIZE / 512] = {0,};
  int i, j;
	 

  p = buf;
  left = size;
  read = 0;
    
  /* here tail means i_size % PAGE_SIZE. So, for blocksize <
	  PAGE_SIZE part of tail may lay in unformatted nodes. Make
	  request to read them */
  i = j = 0;
  while (offset < tail_offset (inode)) {
	 block = offset >> inode->i_sb->s_blocksize_bits;
	 block = reiserfs_bmap(inode, block);
	 if (block) {
		pbh[i] = getblk (inode->i_dev, block, inode->i_sb->s_blocksize);
		i ++;
		requested[j] = 1;
	 }
	 j ++;
	 offset += inode->i_sb->s_blocksize;
	 p += inode->i_sb->s_blocksize;
  }
  if (i)
	 ll_rw_block (READ, i, pbh);


  if (offset == tail_offset (inode)) {
	 /* read direct item(s) */
	 make_cpu_key (&key, inode, offset + 1, TYPE_DIRECT);
    
	 /* p points to the place in the buf where file tail has to be */  
	 while (left) {
		if (search_for_position_by_key (inode->i_sb, &key, &path) == POSITION_NOT_FOUND) {
		  break;
		}

		bh = PATH_PLAST_BUFFER (&path);
		ih = B_N_PITEM_HEAD (bh, PATH_LAST_POSITION (&path));
		pos_in_item = path.pos_in_item;
		chars = ih->ih_item_len - pos_in_item;
		if (chars > left)
		  chars = left;
		
		memcpy (p, B_I_PITEM (bh, ih) + pos_in_item, chars);
		set_cpu_key_k_offset (&key, cpu_key_k_offset (&key) + chars);
		read += chars;
		left -= chars;
		p += chars;
		if (PATH_LAST_POSITION (&path) != B_NR_ITEMS (bh) - 1)
		  /* that was last direct item of the tail */
		  break;
	 }
	 pathrelse (&path);
  } else {
	 /* generic_file_read's readahead request */
  }

  /* finish with unformatted nodes */
  p = buf;
  for (i = j = 0; j < PAGE_SIZE / 512; j ++) {
	 if (requested[j]) {
		wait_on_buffer (pbh[i]);
		memcpy (p, pbh[i]->b_data, pbh[i]->b_size);
		brelse (pbh[i]);
		i ++;
	 }
	 read += inode->i_sb->s_blocksize;
	 left -= inode->i_sb->s_blocksize;
	 p += inode->i_sb->s_blocksize;
  }

  reiserfs_check_path(&path) ;
  return read;
}


#define lock_tail(inode) increment_i_read_sync_counter(inode)
#define unlock_tail(inode) decrement_i_read_sync_counter(inode)


//#if 0

int reiserfs_readpage (struct dentry *dentry, struct page * page) 
{
  struct inode * inode;
    
  inode = dentry->d_inode;
    
  if (has_tail (inode) && 
		tail_offset (inode) < (page->index << PAGE_SHIFT)  + PAGE_SIZE) {
	 /* there is a tail and it is in this page */
	 lock_kernel() ;
	 lock_tail (inode);
	 if (has_tail (inode) && tail_offset (inode) < (page->index << PAGE_SHIFT) + PAGE_SIZE) {
		memset ((char *)page_address (page), 0, PAGE_SIZE);
		read_file_tail (inode, page->index << PAGE_SHIFT, (char *)page_address (page), PAGE_SIZE);
		unlock_tail (inode);
		set_bit (PG_uptodate, &page->flags);
		UnlockPage(page);
		unlock_kernel() ;
		return 0;
	 }
	 unlock_tail (inode);
	 unlock_kernel() ;
  }
  return block_read_full_page (page, reiserfs_get_block);
}

#endif



int reiserfs_prepare_write(struct page *page, unsigned from, unsigned to) {
  return block_prepare_write(page, from, to, reiserfs_get_block) ;
}
int reiserfs_aop_bmap(struct address_space *as, long block) {
  return generic_block_bmap(as, block, reiserfs_get_block) ;
}


#if 0

/* convert tail in unformatted node if it fits to this page */
int reiserfs_writepage (struct dentry * dentry, struct page * page)
{
  struct inode * inode;
  unsigned long block;
  struct buffer_head * bh;
  int unlocked = 0 ;
  loff_t offset = page->index << PAGE_SHIFT;
    
  inode = dentry->d_inode;
    
  lock_kernel() ;
  if (has_tail (inode) && tail_offset (inode) < offset + PAGE_SIZE) {
	 /* there is a tail and it is in this page. get_block will
		 convert tail to the page */
	 if (!page->buffers) {
		reiserfs_warning ("vs-: reiserfs_writepage: page->buffers == 0\n");
		unlock_kernel() ;
		return -EINVAL;
	 }

	 bh = page->buffers;
	
	 while (tail_offset (inode) != offset) {
		offset += inode->i_sb->s_blocksize;
		bh = bh->b_this_page;
	 }

	 block = page->index << (PAGE_SHIFT - inode->i_sb->s_blocksize_bits);
	 unlock_kernel() ;
	 unlocked = 1 ;
	 if (reiserfs_get_block (inode, block, bh, 1))
		return -ENOSPC;
  }

  if (!unlocked) {
	 unlock_kernel() ;
  }
  return block_write_full_page (page, reiserfs_get_block);
}
#endif



#if 0

extern spinlock_t inode_lock;

                                /* This is the problem we are trying to solve: with reiserfs it is a bad thing to defer
                                an update of stat data if the buffer holding the stat data is in RAM. If we defer it,
                                then when we later do the update, the buffer might not still be in RAM.  Consider that
                                inodes have a longer lifetime in RAM, and you can see that this is especially important.

                                Unfortunately, sometimes we need to pass an inode, sometimes an iattr, sometimes just
                                one field, to the function that updates the stat data.  Umpteen different functions that
                                do essentially the same thing.  For now I implement just a straight replacement for
                                mark_inode_dirty -Hans */
                                /* This function is inefficient if only one field of the inode was changed. 
                                 This function can cause schedule.  */
void if_in_ram_update_sd (struct reiserfs_transaction_handle *th, struct inode * inode)
{

  reiserfs_update_stat_data (th, inode, READ_BLOCKS/*DONT_READ_BLOCKS??*/);
  return ;

#ifdef DIRTY_LATER
  mark_inode_dirty(inode);
  return;
#else /* DIRTY_LATER */

  extern struct list_head inode_in_use;


  if (!inode->i_nlink || (S_ISFIFO (inode->i_mode))) { 
    mark_inode_dirty(inode);    /* let us deferr the update since in this case we always follow this with an iput which
                                   will do the update (that is to say, the removal) of the stat data */
    return;
  }

#ifdef REISERFS_CHECK
  if (inode->i_sb) {
#endif /* REISERFS_CHECK */

    /* wait until inode is not locked and then lock it */
    while ((inode->i_state & I_LOCK)) {
      __wait_on_inode(inode);
    }
    inode->i_state |=  I_LOCK;

    if (reiserfs_update_stat_data (th, inode, &path_to_sd, DONT_READ_BLOCKS)) {
      spin_lock(&inode_lock);   /* locks inode lists (not the inode) */
      if ((inode->i_state & I_DIRTY)) {
        struct list_head *insert = &inode_in_use;

        /* mark inode clean and take it off dirty list */
        inode->i_state &= ~I_DIRTY;
#ifdef REISERFS_CHECK
        /* Only add valid (ie hashed) inodes to the in_use list */
        if (!list_empty(&inode->i_hash)) {
#endif /* REISERFS_CHECK */
          list_del(&inode->i_list);
          list_add(&inode->i_list, insert);
#ifdef REISERFS_CHECK
        } 
        else 
          printk("reiser-1805: if_in_ram_update_sd: a dirty inode was not on any inode list\n");
#endif /* REISERFS_CHECK */
      }
      spin_unlock(&inode_lock);
      /*mark_buffer_dirty (bh, 1);*/
    } else {   /* stat data item for this inode no longer in RAM */
      /* if stat data was not found then it must be because it is no longer in RAM, so just mark it
	 dirty for now, and let somebody else write it */
      mark_inode_dirty(inode); 
    }  
    inode->i_state &= ~I_LOCK;
    wake_up(&inode->i_wait);
    pathrelse(&path_to_sd);
#ifdef REISERFS_CHECK
  }
  else
    printk("reiser-1804: if_in_ram_update_sd: !sb, should not happen\n");
#endif /* REISERFS_CHECK */

#endif /* ! DIRTY_LATER */
}


/*#ifdef __KERNEL__*/


#ifndef DIRTY_LATER

/* this generates little if any speedup compared to if_in_ram_update_sd(), might not be worthwhile
   code, probably search_by_key dominates cpu consumption */
void if_in_ram_update_some_sd (struct reiserfs_transaction_handle *th, struct inode * inode,  struct iattr * attr)
{
  struct stat_data * sd;
  INITIALIZE_PATH (path_to_sd);
  struct buffer_head * bh;
  extern struct list_head inode_in_use;
  unsigned int ia_valid = attr->ia_valid;


  if (!inode->i_nlink || (S_ISFIFO (inode->i_mode))) { /* Vladimir, what do you think, is the S_FIFO needed here?-Hans */
    mark_inode_dirty(inode);    /* let us deferr the update since in this case we always follow this with an iput which
                                   will do the update (that is to say, the removal) of the stat data */
    return;
  }

#ifdef REISERFS_CHECK
  if (inode->i_sb) {
#endif /* REISERFS_CHECK */

                                /* wait until inode is not locked and then lock it */
    while ((inode->i_state & I_LOCK)) {
      __wait_on_inode(inode);
    }
    inode->i_state |=  I_LOCK;  /* depends on interrupts never locking inodes, I suppose it is correct.... */
	 
    /* look for the stat data item, if we find it in RAM, sync to it now, else put the inode on the list of dirty inodes
       to be synced someday using mark_inode_dirty and perhaps the sync will happen when the buffer is back in RAM or
       after a bunch of other changes to it have been successfully cached by deferring the I/O */
    if (search_by_key (inode->i_sb, INODE_PKEY (inode), &path_to_sd, DISK_LEAF_NODE_LEVEL, DONT_READ_BLOCKS) == ITEM_FOUND) {
                                /* all of these branches are very inefficient, but the search_by_key is probably the
                                   worst of it */
      bh = PATH_PLAST_BUFFER (&path_to_sd);
      sd = B_N_STAT_DATA (bh, PATH_LAST_POSITION (&path_to_sd));
		if (ia_valid & ATTR_MODE)
		  sd->sd_mode = cpu_to_le16 (inode->i_mode);
		if (ia_valid & ATTR_UID)
		  sd->sd_uid = cpu_to_le16 (inode->i_uid);
		if (ia_valid & ATTR_GID)
		  sd->sd_gid = cpu_to_le16 (inode->i_gid);
		if (ia_valid & ATTR_SIZE)
		  sd->sd_size = cpu_to_le32 (inode->i_size);
		if (ia_valid & ATTR_MTIME)
		  sd->sd_mtime = cpu_to_le32 (inode->i_mtime);
		if (ia_valid & ATTR_CTIME)
		  sd->sd_ctime = cpu_to_le32 (inode->i_ctime);
      spin_lock(&inode_lock);   /* locks inode lists (not the inode) */
      if ((inode->i_state & I_DIRTY)) {
        struct list_head *insert = &inode_in_use;
		  
        /*      mark inode clean and take it off dirty list */
        inode->i_state &= ~I_DIRTY;
#ifdef REISERFS_CHECK
        /* Only add valid (ie hashed) inodes to the in_use list */
        if (!list_empty(&inode->i_hash)) {
#endif /* REISERFS_CHECK */
          list_del(&inode->i_list);
          list_add(&inode->i_list, insert);
#ifdef REISERFS_CHECK
        }
        else 
          printk("reiser-1806: if_in_ram_update_some_sd: a dirty inode was not on any inode list, maybe a pipe?");
#endif /* REISERFS_CHECK */
      }
      spin_unlock(&inode_lock);
      journal_mark_dirty (th, inode->i_sb, bh);
    } else {   /* stat data item for this inode no longer in RAM */
      inode->i_state &= ~I_LOCK; /* probably should combine two lines that unlock inode by postponing the unlock.... */
                                /* if inode was not found then it must be because it is no longer in RAM, so just
                                   mark it dirty for now, and let somebody else write it */
      mark_inode_dirty(inode); 
    }  
    inode->i_state &= ~I_LOCK;
    wake_up(&inode->i_wait);
    pathrelse(&path_to_sd);
#ifdef REISERFS_CHECK
  }
  else
    printk("reiser-1807: if_in_ram_update_some_sd: !sb, should not happen");
#endif /* REISERFS_CHECK */
}

#endif


void inline reiserfs_inode_setattr(struct reiserfs_transaction_handle *th, struct inode * inode, struct iattr * attr)
{
  unsigned int ia_valid = attr->ia_valid;
  /* struct path path_to_sd; */

  if (ia_valid & ATTR_UID)
	 inode->i_uid = attr->ia_uid;
  if (ia_valid & ATTR_GID)
	 inode->i_gid = attr->ia_gid;
  if (ia_valid & ATTR_SIZE)
	 inode->i_size = attr->ia_size;
  if (ia_valid & ATTR_ATIME)
	 inode->i_atime = attr->ia_atime;
  if (ia_valid & ATTR_MTIME)
	 inode->i_mtime = attr->ia_mtime;
  if (ia_valid & ATTR_CTIME)
	 inode->i_ctime = attr->ia_ctime;
  if (ia_valid & ATTR_MODE) {
	 inode->i_mode = attr->ia_mode;
	 if (!in_group_p(inode->i_gid) && !capable(CAP_FSETID))
		inode->i_mode &= ~S_ISGID;
  }
#ifdef DIRTY_LATER
  mark_inode_dirty(inode) ;  
  /* reiserfs_update_stat_data (inode, &path_to_sd, READ_BLOCKS); 
	  pathrelse (&path_to_sd); */
#else /* note, this code is broken and MUST not be used */
  if_in_ram_update_some_sd (th, inode, attr);
#endif
}


int reiserfs_notify_change(struct dentry * dentry, struct iattr * attr)
{
  struct inode *inode = dentry->d_inode;
  struct reiserfs_transaction_handle th ;
  int error;

  /* I'm cheating here.  reiserfs_inode_setattr does not make journal calls with
  ** dirty later turned on.  Dirty later must be on for now, so I'm not doing
  ** a journal_begin here.
  */
  th.t_trans_id = 0 ; 
  error = inode_change_ok(inode, attr);
  if (!error)
	 reiserfs_inode_setattr(&th, inode, attr);
  return error;
}

/* I believe that further optimizing this code is best done by optimizing search_by_key.. */

/*#endif*/ /* __KERNEL__ */


#endif
