/*
 * Copyright 1996, 1997, 1998 Hans Reiser, see reiserfs/README for licensing and copyright details
 */
#ifdef __KERNEL__

#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/reiserfs_fs.h>
#include <asm/uaccess.h>
#include <linux/smp_lock.h>

#else

#include "nokernel.h"

#endif


static int reiserfs_readlink(struct dentry *, char *, int);
static struct dentry * reiserfs_follow_link(struct dentry *, struct dentry *, unsigned int);


struct inode_operations reiserfs_symlink_inode_operations = {
    readlink:	reiserfs_readlink,
    follow_link:	reiserfs_follow_link,
};


/* readlink does not append a NUL character to buf (from man 2 readlink) */
static void read_symlink (struct inode * inode, char * buf, int memmode)
{
    struct cpu_key key;
    int pos_in_item, chars, len = ROUND_UP (inode->i_size);
    INITIALIZE_PATH (path);
    struct item_head * ih;

    make_cpu_key (&key, inode, 1, TYPE_DIRECT, 3);

    while (len) {
	if (search_for_position_by_key (inode->i_sb, &key, &path) == POSITION_NOT_FOUND)
	    reiserfs_panic (inode->i_sb, "vs-17000: read_symlink: symlink item not found");
	ih = PATH_PITEM_HEAD(&path);
	pos_in_item = path.pos_in_item;
	chars = ih->ih_item_len - pos_in_item;
	if (memmode == REISERFS_KERNEL_MEM)
	    memcpy (buf, B_I_PITEM(PATH_PLAST_BUFFER(&path),ih) + pos_in_item, chars);
	else
	    copy_to_user (buf, B_I_PITEM(PATH_PLAST_BUFFER(&path),ih) + pos_in_item, chars);
	buf += chars;
	set_cpu_key_k_offset (&key, cpu_key_k_offset (&key) + chars);
	len -= chars;

#ifdef REISERFS_CHECK
	if (len < 0)
	    reiserfs_panic (inode->i_sb, "vs-17005: read_symlink: "
			    "too many bytes read from symlink (%d). Must be %d",
			    inode->i_size - len, inode->i_size);
#endif

    }

    decrement_counters_in_path(&path);
}


static struct dentry * reiserfs_follow_link (struct dentry * dentry, struct dentry * base, unsigned int follow)
{
  struct inode * inode = dentry->d_inode;
  char * buf;

  reiserfs_check_lock_depth("follow_link") ;

  buf = reiserfs_kmalloc (inode->i_size + 1, GFP_KERNEL, inode->i_sb);
  if (buf == 0) {
    dput (base);
    return ERR_PTR(-ENOMEM);
  }

  read_symlink (inode, buf, REISERFS_KERNEL_MEM);
  buf[inode->i_size] = 0;

  UPDATE_ATIME(inode);
  base = lookup_dentry (buf, base, follow);
  reiserfs_kfree (buf, inode->i_size + 1, inode->i_sb);

  return base;
}


static int reiserfs_readlink (struct dentry * dentry, char * buffer, int buflen)
{
  struct inode * inode = dentry->d_inode;
  
  reiserfs_check_lock_depth("readlink") ;
  read_symlink (inode, buffer, REISERFS_USER_MEM);
  return inode->i_size;
}





